#!/bin/perl
#################################################################################
# Created       : Martin Foster
# Modified      : 27/12/2002
#################################################################################
#
# System Configuration - Harmonized script designed to behave as a Control Panel
# Copyright (C) 2000-2002  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 204 - 817, 5th Street NE
#                 Calgary, Alberta
#                 T2E 3W9
#
#################################################################################
use CGI qw(:standard);						# Common gateway interface
use CGI::Carp qw(fatalsToBrowser);				# CGI Error logs
use Digest::MD5 qw(md5_hex);					# Digest handling
use strict;							# Strict variable enforcement

use Ethereal::Audit;						# Audit trail
use Ethereal::Database;						# Database handler
use Ethereal::Date;						# Date handler
use Ethereal::Filter;						# Filter handler
use Ethereal::Home;						# Homepage handler
use Ethereal::Login;						# Login functionality
use Ethereal::Mail;						# Mail handler
use Ethereal::Param;						# Parameter control
use Ethereal::Table;						# Table handle

#################################################################################
# Data Members
#################################################################################
my $cgi;							# Common gateway interface handle
my $database;							# Database handle
my $login;							# Login handle
my $param;							# Parameter handle
my $table;							# Table name

my $address;							# Script address
my $inline;							# Inline address
my $scriptname;							# Script name


# Script addresses and names
my %scripts = (
	'Access'    => 'access.pl',
	'Attribute' => 'attrib.pl',
	'Contact'   => 'contact.pl',
	'Font'      => 'font.pl',
	'Gummy'     => 'gummy.pl',
	'Template'  => 'template.pl',
	'Realm'     => 'realm.pl',
	'User'      => 'system.pl',
	'Weblog'    => 'weblog.pl'
);

# System hash
my %config;


#################################################################################
# Program Area
#################################################################################

	# Initial handle
	$cgi      = new CGI;
	$database = new Ethereal::Database();
	$table    = new Ethereal::Table($cgi);

	# Retrieve parameters
	my @allow = $cgi->param('ALLOW');
	my @block = $cgi->param('BLOCK');

	# Connect and fetch
	$database->Connect();

	# Parameter handling
	$param = new Ethereal::Param($database, $cgi);
	$param->GetParam();


	# Document header
	print $cgi->header();
	$database->DocumentGetHeader();

	# Login
	$login = new Ethereal::Login($database, $cgi, $param);

	# Authenthication
	if ($login->GetVerificationAdmin())
	{
		# Retreive script
		$scriptname = $cgi->url(-relative=>1);

		# Generate widget
		$inline = $param->EmbedInline('USER', $param->{'USER'}, 'CRYPT', $param->{'CRYPT'});


		# Menu control
		# Preperation
		print "\n<CENTER>\n";

		# Loop
		foreach $address (sort(keys(%scripts)))
		{
			# Print link
			if ($scripts{$address} eq $scriptname)
			{ print "[<A HREF=\"$scripts{$address}$inline\"><I><FONT SIZE=3>$address</FONT></I></A>]"; }

			else
			{ print "[<A HREF=\"$scripts{$address}$inline\"><B><FONT SIZE=3>$address</FONT></B></A>]"; }
		}

		# Finish
		print "\n</CENTER>\n<BR>\n";


		# Access administration
		if ($scriptname =~ /^$scripts{'Access'}/i)
		{
			# Logic Tree
			# Access Control
			if (defined($param->{'ROOM'}))
			{ AccessControl($database, $cgi, $param, \@allow, \@block); }

			# Realm Listing
			else
			{ RealmSelect($database, $cgi, $param); }
		}

		# Attribute administrator
		elsif ($scriptname =~ /^$scripts{'Attribute'}/i)
		{
			# Form start
			print $cgi->start_form();
			print $param->EmbedNormal('USER', $param->{'USER'}, 'CRYPT', $param->{'CRYPT'}), "\n";

			# List data
			DataCycler($database, $cgi, $param);

			# Form end
			print $cgi->p($cgi->submit(), $cgi->reset()), "\n<br>\n";
			print $cgi->end_form(), "\n";
		}

		# Contact elements
		elsif ($scriptname =~ /^$scripts{'Contact'}/i)
		{
			# Handler
			ContactHandler($database, $cgi, $param, $table);
		}


		# Document administration
		elsif ($scriptname =~ /^$scripts{'Template'}/i)
		{
			# If no DOCTYPE then list options 
			DocumentSelect($database, $cgi, $param) if (!defined($param->{'DOCTYPE'}));			
	
			# If only DOCTYPE is specified then allow to modify
			if ((defined($param->{'DOCTYPE'})) && (!defined($param->{'DOCDATA'})))
			{
				DocumentSelect($database, $cgi, $param);
				DocumentModify($database, $cgi, $param); 		
			}

			# If only DOCTYPE and DOCDATA are specified
			elsif ((defined($param->{'DOCTYPE'})) && (defined($param->{'DOCDATA'})))
			{
				DocumentSelect($database, $cgi, $param);
				DocumentModify($database, $cgi, $param); 		
			}
		}

		# Font handler
		elsif ($scriptname =~ /^$scripts{'Font'}/i)
		{
			# Font handling subsystem
			FontHandler($database, $cgi, $param, $table);
		}

		# Gummy handler
		elsif ($scriptname =~ /^$scripts{'Gummy'}/i)
		{
			# Class selection
			unless (defined($param->{'CLASS'}))
			{
				# Selection of class
				GummySelect($database, $cgi, $param, $table);
			}

			else
			{
				# Handle addition and removal
				GummyHandler($database, $cgi, $param, $table);
			}
		}

		# Realm administration
		elsif ($scriptname =~ /^$scripts{'Realm'}/i)
		{
			# Realm additions
			if (defined($param->{'ADDITION'}))
			{ RealmAddition($database, $cgi, $param); }

			# Realms displays
			elsif (defined($param->{'REALMTYPE'}))
			{ RealmDisplay($database, $cgi, $param); }

			# Selection
		 	else 
			{ RealmType($database, $cgi, $param); }
		}

		# Weblog administration
		elsif ($scriptname =~ /^$scripts{'Weblog'}/i)
		{
			# Weblog handling system
			WeblogHandler($database, $cgi, $param, $table);
		}

		# User administration
		else
		{
			# Handles the primary modifications of a user
			if (defined($param->{'LOGIN'}))
			{
				UserModify($database, $cgi, $param);
			}

			# Basic search with user selection and removal
			elsif (defined($param->{'SEARCH'}))
			{
				UserSelect($database, $cgi, $param);
			}

			# Basic Prompt
			else
			{ 
				UserSearch($database, $cgi, $param);
			}		
		}
	}

	# Document footer
	$database->DocumentGetFooter();


#################################################################################
# Sub-Routines
#################################################################################

#####################
# Access Control
#
# Access Control Lists is essentially the role of this script.  This specific 
# subroutine is aimed speficially at handling who can access which Realm
# and perform supervisory functions.

sub AccessControl
{
	#####################
	# Data Members

	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter list
	my $allow    = shift;					# Access granted list
	my $block    = shift;					# Access denied list
	
	my $data     = $database->{'HANDLE'};			# Database connection handle

	my $audit;						# Audit handle
	my $handle;						# Handle name
	my $instance;						# Instances

	my $badd     = '>>';					# Addition button
	my $brem     = '<<';					# Remove button

	my $inline;						# Inline parameter list
	my $url      = $cgi->url(-absolute=>1);			# Self referencing link

	my @allow;						# Allowed
	my @block;						# Blocked


	#####################
	# Program Files

	# Audit handle
	$audit = new Ethereal::Audit($database, $cgi, $param);

	# Display realm
	print $cgi->h2($param->{'ROOM'});

	# Data handling 
	# Additions
	if (defined($param->{$badd}))
	{
		# Special additions
		if ((defined($param->{'NALLOW'})) 
		 && (length($param->{'NALLOW'}) > 2))
		{
			# Determine validity
			($param->{'NALLOW'}) = $database->DataGet("SELECT PuppeteerLogin,
				  PuppeteerPassword
				FROM Puppeteer
				WHERE PuppeteerLogin=?", $param->{'NALLOW'});

			# If user exists
			if (defined($param->{'NALLOW'}))
			{
				# Adjust privs
				$data->do("UPDATE Puppeteer
					SET   PuppeteerSuperPrivs='yes',
					      PuppeteerModPrivs='no'
					WHERE PuppeteerLogin=?", {},

				 $param->{'NALLOW'});

				# Add to be added normally
				unshift(@{$block}, $param->{'NALLOW'});
			} 
		}

		# Handle addition
		foreach $handle (@{$block})
		{
			# Insert Entry
			$data->do("INSERT INTO RealmSupervisor 
				(RealmName,
				 PuppeteerLogin)
				VALUES (?,?)", {},

			 $param->{'ROOM'},
			 $handle);

			# Audit table
			$audit->AccessAdd($param->{'ROOM'}, $handle);	
		}

		# Cleanup	
		$param->Cleanup($badd);
	}

	# Removal
	elsif (defined($param->{$brem}))
	{
		foreach $handle (@{$allow})
		{
			# Determin amount of instances
			($instance) = $database->DataGet("SELECT COUNT(*)
				FROM RealmSupervisor
				WHERE PuppeteerLogin=?", $handle);

			# Remove priviliges to all those that exist once
			if ((defined($instance)) && ($instance == 1))
			{
				$data->do("UPDATE Puppeteer
					SET   PuppeteerSuperPrivs='no',
					PuppeteerModPrivs='no'
					WHERE PuppeteerLogin=?", {}, $handle);
			}



			# Remove data
			$data->do("DELETE FROM RealmSupervisor
				WHERE PuppeteerLogin=?
				  AND RealmName=?", {},

				$handle,
				$param->{'ROOM'});

			# Audit table
			$audit->AccessAdd($param->{'ROOM'}, $handle);	
		}

		# Cleanup
		$param->Cleanup($brem);
	}
	


	# Data retreival
	# Retrieve variables
	$database->GetList(\@block, "SELECT PuppeteerLogin 
		FROM  Puppeteer 
		WHERE PuppeteerSuperPrivs='yes' 
		ORDER BY PuppeteerLogin"); 
	$database->GetList(\@allow, "SELECT p.PuppeteerLogin 
		FROM  Puppeteer p, RealmSupervisor r 
		WHERE p.PuppeteerLogin = r.PuppeteerLogin
		  AND r.RealmName=?
		  AND p.PuppeteerSuperPrivs='yes'
		ORDER BY p.PuppeteerLogin", $param->{'ROOM'});

	# Generate inline
	($inline) = $param->EmbedInline('USER', $param->{'USER'}, 'CRYPT', $param->{'CRYPT'}); 


	# Form display
	# Form start
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";

	# Display
	print "<CENTER><TABLE WIDTH=\"500\">\n",
	      "\t<TR>\n\t\t<TD COLSPAN=\"3\"><CENTER><B><A HREF=\"$url$inline\">[Back]</A></B></CENTER></TD>\n\t</TR>\n",
	      "\t<TR>\n",
	      "\t\t<TD><B>Supervisors</B></TD>",
	      "\t\t<TD WIDTH=\"50\">&nbsp;</TD>",
	      "\t\t<TD><B>Realm Supervisors</B></TD>",
	      "\t</TR>\n",
	      "\t<TR>\n",
	      "\t\t<TD>", $cgi->scrolling_list('BLOCK', \@block, $block[0], '5', 'true'), "</TD>",
	      "\t\t<TD WIDTH=\"50\">", $cgi->submit($badd), "<BR><BR>", $cgi->submit($brem), "</TD>",
	      "\t\t<TD>", $cgi->scrolling_list('ALLOW', \@allow, $allow[0], '5', 'true'), "</TD>",
	      "\t</TR>\n",
	      "\t<TR>\n",
	      "\t\t<TD>", $cgi->textfield('NALLOW', '', 15, 10), "</TD>",
	      "\t\t<TD WIDTH=\"50\">&nbsp;</TD>",
	      "\t\t<TD>&nbsp;</TD>",
	      "</TABLE></CENTER>";

	# Form end
	print $cgi->end_form(), "\n";
}


#####################
# Contact Handler
#
# Designed to show the administrator all contact information that can be gathered
# and allow for it to be removed.

sub ContactHandler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI filehandle
	my $param    = shift;					# Hash of parameters					

	my $audit;						# Auditing handle
	my $table    = shift;						# Auditing table

	my $data     = $database->{'HANDLE'};			# Direct handle
	my $statement;						# Database statement

	my $contact;						# Contact handle
	my $name;						# Contact name
	my $type;						# Contact type

	my $filter;						# Filter widget
	my @filter;						# Filter types


	#####################
	# Program area

	# Define instances
	$audit = new Ethereal::Audit($database, $cgi, $param);

	# Initial assigment
	$type = 'NONE';

	# Filter types
	@filter = qw(None Email HTTP Numeric);
	$filter = $cgi->radio_group('FILTER', \@filter, $filter[0], 'true');


	# Add entry
	if (defined($param->{'Addition'}))
	{
		# Insertion
		if ((defined($param->{'CONTACT'}))
		 && (length($param->{'CONTACT'}) > 2))
		{
			# Assignment
			$param->{'TYPE'} = (length($param->{'TYPE'}) > 2) ? $param->{'TYPE'} : undef;


			$data->do("INSERT INTO Contact
				(ContactName,
				 ContactType,
				 ContactFilter)
				VALUES(?,?,?)", {},

			 $param->{'CONTACT'},
			 lc($param->{'TYPE'}),
			 lc($param->{'FILTER'}));

			# Audit entry
			$audit->ContactAdd($param->{'CONTACT'});
		}

		# Cleanup
		$param->Cleanup('Addition');
	}



	# Display handles
	# Prepare and execute
	$statement = $data->prepare("SELECT
		 ContactName,
		 ContactType,
		 ContactFilter
		FROM Contact
		ORDER BY ContactType, ContactName");
	$statement->execute();


	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat());

	# Table top
	$table->MakeTop();
	$table->MakeSingle('<FONT SIZE=3><B>Contact elements</B></FONT>');


	# Loop and display
	while ($contact = $statement->fetchrow_hashref())
	{
		# Name assignment
		$name =  $contact->{'ContactName'};
		$name =~ s/\W//s;


		# Removal
		if (defined($param->{$name}))
		{
			# Remove
			# Main table
			$data->do("DELETE FROM Contact
				WHERE ContactName=?", {},
			 $contact->{'ContactName'});

			# Contact table
			$data->do("DELETE FROM PuppeteerContact
				WHERE ContactName=?", {},
			 $contact->{'ContactName'});


			# Audit entry
			$audit->ContactRem($contact->{'ContactName'});

			# Cleanup
			$param->Cleanup($name);
		}

		# Display
		else
		{
			# Type display
			# Ensure assignment
			$contact->{'ContactType'} = (defined($contact->{'ContactType'})) ? $contact->{'ContactType'} : 'none';

			# Determine need
			if (uc($contact->{'ContactType'}) ne $type)
			{
				# Massage
				$type = uc($contact->{'ContactType'});

				# Display
				$table->MakeBlank();
				$table->MakeSingle("<B>$type<B>");
			}	


			# Standard display
			$table->MakeValid(
				"<B>$contact->{ContactName}<B> (<I>$contact->{ContactFilter}</I>)",
				$cgi->checkbox(-name=>$name, -label=>'')
			 );
		}
	}

	# Finish Query
	$statement->finish();

	# Submit
	$table->MakeBlank();
	$table->MakeValid('&nbsp;', $cgi->submit('Remove'));


	# Spacer
	$table->MakeBlank();
	$table->MakeBlank();
	$table->MakeBlank();


	# Addition entries
	$table->MakeSingle('<FONT SIZE=3><B>Addition</B></FONT>');

	# Widgets
	$table->MakeValid('<B>Contact Name</B>', $cgi->textfield('CONTACT', '', 15, 25));
	$table->MakeValid('<B>Type</B>', $cgi->textfield('TYPE', '', 10, 10));
	$table->MakeValid('<B>Filter type</B>', $filter);
	$table->MakeBlank();
	$table->MakeValid('&nbsp;', $cgi->submit('Addition'));


	# Finish up
	# End table
	$table->MakeBottom();

	# End form
	print $cgi->end_form()
}


#####################
# Data Cycler
#
# A simple system designed to make the system more adaptable to new sections 
# being added to the system.   

sub DataCycler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI filehandle
	my $param    = shift;					# Hash of parameters					
	my $choice   = shift;					# Choice of hash to open

	my @sections;						# List of sections
	my $section;						# Singular section


	#####################
	# Program area

	# Retreive sections
	$database->GetList(\@sections, 'SELECT DISTINCT(TagType) FROM Tag');


	# Cycle through sections
	foreach $section (sort @sections)
	{
		DataHandler($database, $cgi, $param, $section);
	}
}


#####################
# Data Handler
#
# Takes in a CGI handle and the name of the hash to activate and displays 
# information within those hashes.   If changes, will make the change neccesary.

sub DataHandler							# CGI, Number for hash
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI filehandle
	my $param    = shift;					# Hash of parameters					
	my $dbm	     = shift;					# Hash being used

	my $audit;						# Audit handle
	my $db       = $database->{'HANDLE'};			# Database connection handle

	my $choice;						# Current selection

	my $key;						# Arbitrary key
	my $value;						# Arbitrary value
	my $data;						# Data of hash

	my %hash;						# Arbitrary hash


	#####################
	# Program area

	# Audit handle
	$audit = new Ethereal::Audit($database, $cgi, $param);

	# Link with hash
	$database->GetHash(\%hash, $dbm);

	# Display
	DisplayHandlerHeading($cgi, $dbm);
	DisplayHandlerTop($cgi, $dbm);
	DisplayHandlerTitle($cgi, $dbm);	

	# Loop through and display information
	foreach $choice (sort keys %hash)
	{
		# If diffrent change
		if ((defined($param->{"$dbm$choice"})) && ($param->{"$dbm$choice"} ne $hash{$choice})) 
		{
			$data = $param->{"$dbm$choice"};

			# Update values
			$hash{$choice} = $data;

			$db->do("UPDATE Tag
				SET TagData=?
				WHERE TagName=?
				  AND TagType=?", {},

				 $data,
				 $choice,
				 $dbm);

			# Display fact of change
			DisplayHandlerNewInfo($cgi, $dbm, $choice, '*', $hash{$choice});

			# Audit Trail
			$audit->TagChange($choice);
		}
			
		# Else display as no one really cares
		else
		{
			DisplayHandlerInfo($cgi, $dbm, $choice, '&nbsp;', $hash{$choice});
		}
	}
	DisplayHandlerBottom($cgi, $dbm);
}


#####################
# Display Handler
#
# Make formatted tables based on information received and options. 

sub DisplayHandler
{
	#####################
	# Data members
	my $cgi    = shift;					# CGI filehandle
	my $choice = shift;					# Choice of hash to open
	my $dbm    = shift;					# DBM being worked on

	my $widget;						# Widget type
	my @title = qw(Key Value);				# Titles



	#####################
	# Program area

	# Top of Table
	if ($choice == 0)					# Table top
	{  
		print "<TABLE COLS=3 WIDTH=\"100%\">\n";
	}

	# Title section of table
	elsif ($choice == 1)					# Table title
	{
		print "\t<TR>\n";				# Row beginning

		# Print first three title columns
		print "\t\t<TD WIDTH=\"200\"><B><FONT SIZE=+1>$title[0]</FONT></B></TD>\n";
		print "\t\t<TD WIDTH=\"10\">&nbsp;</TD>\n";
		print "\t\t<TD><FONT SIZE=+1><B>$title[1]</FONT></B></TD>\n";		

		print "\t</TR>\n";				# Row end
	}

	# Information row of table
	elsif ($choice == 2)					# Table data
	{
		print "\t<TR>\n";				# Row beginning

		# Widget selection
		$widget = ($_[0] =~ /^Tmpl/) ? 
			$cgi->textarea(-name=>"$dbm$_[0]", -rows=>5, -columns=>85, -wrap=>'OFF', -default=>$_[2]) :
			$cgi->textfield("$dbm$_[0]",$_[2],80,4000);

		# Print three columns of data
		print "\t\t<TD><B>$_[0]</B></TD>\n";
		print "\t\t<TD><FONT COLOR=\"\#00CC00\">$_[1]</FONT></TD>\n";
		print "\t\t<TD>", $widget, "</TD>\n";
		
		print "\t</TR>\n";				# Row end
	}

	# Information row of table
	elsif ($choice == 3)					# Table data
	{
		print "\t<TR>\n";				# Row beginning

		# Widget selection
		$widget = ($_[0] =~ /^Tmpl/) ? 
			$cgi->textarea(-name=>"$dbm$_[0]", -rows=>5, -columns=>85, -wrap=>'OFF', -default=>$_[2]) :
			$cgi->textfield("$dbm$_[0]",$_[2],80,4000);

		# Print three columns of data
		print "\t\t<TD><B>$_[0]</B></TD>\n";
		print "\t\t<TD><FONT COLOR=\"\#FF0000\">$_[1]</FONT></TD>\n";
		print "\t\t<TD>", $widget, "</TD>\n";

		print "\t</TR>\n";				# Row end
	}

	# Bottom of table
	elsif ($choice == 4)					# Table bottom
	{
		print "</TABLE>\n\n<BR>";
	}

	# Heading of table
	elsif ($choice == 5)					# Table bottom
	{
		# Arrange title
		$dbm =~ tr/[a-z]/[A-Z]/;

		# Print title
		print "<BR>\n<BR>\n<B><FONT SIZE=+2>$dbm</FONT></B>\n<BR>\n";
	}
}

# Alias table top
sub DisplayHandlerTop
{ DisplayHandler(shift, 0, shift); }

# Alias table bottom
sub DisplayHandlerBottom
{ DisplayHandler(shift, 4, shift); }

# Alias table title
sub DisplayHandlerTitle
{ DisplayHandler(shift, 1, shift); }

# Alias table info
sub DisplayHandlerInfo
{ DisplayHandler(shift, 2, shift, @_); }

# Alias table new info
sub DisplayHandlerNewInfo
{ DisplayHandler(shift, 3, shift, @_); }

# Alias table heading
sub DisplayHandlerHeading
{ DisplayHandler(shift, 5, shift); }


#####################
# Document Modify
#
# Simply modifies documents available and allows one to select them.

sub DocumentModify						# Database, CGI
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter list

	my $data     = $database->{'HANDLE'};			# Database connection handle

	my $audit;						# Audit handle
	my $document;						# Document
	

	#####################
	# Program area	
	
	# Audit handle
	$audit = new Ethereal::Audit($database, $cgi, $param);

	# If defined save
	if (defined($param->{'DOCDATA'}))
	{
		$data->do("UPDATE WebchatText 
			SET WebchatTextContent=? 
			WHERE WebchatTextName=?", {},  

			$param->{'DOCDATA'}, 
			$param->{'DOCTYPE'});

		# Audit Trail
		$audit->Document($param->{'DOCTYPE'});

	}

	# Retrieve contents
	($document) = $database->DataGet("SELECT WebchatTextContent
		FROM WebchatText
		WHERE WebchatTextName=?", $param->{'DOCTYPE'});

	# Cleanup
	$param->Cleanup('DOCDATA');


	# Start from
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";	

	# Display HTML and information
	print "\t<B>Opened document: $param->{DOCTYPE}</B>\n\t<BR>\n";
	print "\t", $cgi->textarea(-name=>'DOCDATA',-rows=>25,-columns=>85,-wrap=>'OFF',-default=>$document), "\n";


	# End of form
	print "\t", $cgi->p($cgi->submit('Save Changes')), "\n";
	print $cgi->end_form(), "\n";
}


#####################
# DocumentSelect
#
# Simply lists documents available and allows one to select them.

sub DocumentSelect						# Database, CGI
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter list

	my $list;						# Default list unit
	my @list;						# List of fields


	#####################
	# Program area

	# Retrieve list of documents
	$database->GetListDocuments(\@list);

	# Determin defaults
	$list = (defined($param->{'DOCTYPE'})) ? $param->{'DOCTYPE'} : $list[0];	


        # Form data		
	print $cgi->start_form();
	print $param->EmbedNormal('USER', $param->{'USER'}, 'CRYPT', $param->{'CRYPT'}), "\n";

	# Prompt and Display Document choices
	print "\t", $cgi->p('Document list: ', $cgi->popup_menu('DOCTYPE', \@list, $list)), "\n";
	print "\t", $cgi->p($cgi->submit('Open Document')), "\n";


	print $cgi->end_form(), "\n";				# Ending HTML form
}


#####################
# FontHandler
#
# Will simply display the font colour, face and size as to allow the administrator
# to add and remove them as need be.   Similiar in function to the gummy handler.

sub FontHandler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter reference
	my $table    = shift;					# Table parameter

	my $audit;						# Auditing handle
	my $data     = $database->{'HANDLE'};			# Database handle

	my $res;						# Results			
	my $statement;						# Database statement

	my $name;						# Font name

	my $ctype;						# Current working type
	my $type;						# Font type


	#####################
	# Program area

	# Audit handle
	$audit = new Ethereal::Audit($database, $cgi, $param);


	# Additions
	if ((defined($param->{'NAME'}))
	 && (defined($param->{'CODE'}))
	 && (length($param->{'NAME'}) > 0))
	{
		# Force lower case
		$param->{'TYPE'} = lc($param->{'TYPE'});

		# Insert information
		$data->do("INSERT INTO Font
			(FontName,
			 FontType,
			 FontCode)
			VALUES (?,?,?)", {},

		 $param->{'NAME'},
		 $param->{'TYPE'},
		 $param->{'CODE'});


		# Audit
		$audit->FontRem($param->{'TYPE'},$param->{'NAME'});

		# Cleanup
		$param->Cleanup('CODE','NAME','TYPE');
	}



	# Query management
	# Prepare and execute
	$statement = $data->prepare("SELECT 
		 FontName,
		 FontType,
		 FontCode
		FROM  Font
		ORDER BY FontType, FontName");
	$statement->execute();


	# Form handle
	print $cgi->start_form();
	print $param->EmbedNormal('USER', $param->{'USER'}, 'CRYPT',  $param->{'CRYPT'});

	# HTML display
	# Table Top
	print "<CENTER>\n<TABLE WIDTH=500 CELLPADDING=5>\n";

	# Cycle through gummies
	while ($res = $statement->fetchrow_hashref())
	{
		# Title
		if ((!defined($type))
		 or ($type ne $res->{'FontType'}))
		{
			# Uppercase
			$type = uc($res->{'FontType'});


			# Display row
			print " <TR>\n",
			      "  <TD COLSPAN=3><BR>\n<BR>\n<B><FONT SIZE=+2>$type</FONT></B><BR></TD>\n",
			      " </TR>\n";


			# Alternate for next pass
			$type = $res->{'FontType'};
		}

		# Name handling
		$name = $res->{FontName};
		$name =~ s/\s//gs;

		# Handle removal
		if (defined($param->{$name}))
		{
			# Remove gummy
			$data->do("DELETE FROM Font
				WHERE FontName=?
				  AND FontType=?", {},

			 $res->{'FontName'},
			 $res->{'FontType'});


			# Audit action
			$audit->FontRem($res->{'FontType'},$res->{'FontName'});

			# Cleanup
			$param->Cleanup($name);
		}

		# Display if not removed
		else
		{
			# Display row
			if ($type eq 'color')
			{
				# Display colour
				print " <TR>\n",
				      "  <TD><B>$res->{FontName}</B></TD>\n",
				      "  <TD BGCOLOR=\"$res->{FontCode}\" WIDTH=15>&nbsp;</TD>\n",
				      "  <TD><CENTER>", $cgi->checkbox(-name=>$name, -label=>''), "</CENTER></TD>\n",
				      " </TR>\n";
			}
			else
			{
				# Current working type
				$ctype = uc($type);

				# Simply show a blank space
				print " <TR>\n",
				      "  <TD><FONT $ctype=\"$res->{FontCode}\">$res->{FontName}</FONT></TD>\n",
				      "  <TD>&nbsp;</TD>\n",
				      "  <TD><CENTER>", $cgi->checkbox(-name=>$name, -label=>''), "</CENTER></TD>\n",
				      " </TR>\n";
			}
		}
	}

	# Display row
	print " <TR>\n",
	      "  <TD>&nbsp;</TD>\n",
	      "  <TD>&nbsp;</TD>\n",
	      "  <TD><CENTER>", $cgi->submit('Remove'), "</CENTER></TD>\n",
	      " </TR>\n";


	# End of table
	print "</TABLE>\n</CENTER>\n";
	
	# End query
	$statement->finish();


	# Widget
	my @radiotype  = qw(Color Face Size); 
	my $radiofield = $cgi->radio_group('TYPE', \@radiotype, $radiotype[0], 'true');


	# Spacer
	print "<BR>\n<BR>\n";

	# Form handle
	# Addition section
	# Table top
	$table->MakeTop();
	$table->MakeSingle("<B><FONT SIZE=+2>ADDITION</FONT></B>");
	$table->MakeBlank();

	# Display rows
	$table->MakeValid('<B>Name</B>',  $cgi->textfield('NAME', '', 20, 20));
	$table->MakeValid('<B>Code</B>', $cgi->textfield('CODE', '', 20, 35));
	$table->MakeBlank();

	$table->MakeValid('<B>Type</B>', $radiofield);
	$table->MakeBlank();

	$table->MakeSingle($cgi->submit('Add'));
	$table->MakeBottom();
	
	
	# End form
	print $cgi->end_form(), "\n";
}


#####################
# GummyHandler
#
# Very simple sub-routine that will simply allow the user to select the class
# of gummies to search for.

sub GummyHandler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter reference
	my $table    = shift;					# Table parameter

	my $audit;						# Auditing handle
	my $data     = $database->{'HANDLE'};			# Database handle

	my $res;						# Results			
	my $statement;						# Database statement

	my $ident;						# Gummy indentier
	my $group    = 'None';					# Gummy grouping 
	my $name;						# Name of gummy
	my $realm    = 'system';				# Realm search


	#####################
	# Program area

	# Audit handle
	$audit = new Ethereal::Audit($database, $cgi, $param);


	# Addition
	# Make sure gummy is defined
	if ((defined($param->{'GUMMY'}))
	 && (defined($param->{'GPATH'}))
	 && (length($param->{'GUMMY'}) > 2))
	{
		# Group handler
		$param->{'GROUP'} = lc($param->{'GROUP'});
		$param->{'GROUP'} = (length($param->{'GROUP'}) > 2) ? $param->{'GROUP'} : undef;


		# Create entry
		$data->do("INSERT INTO Gummy
			(GummyName,
			 GummyClass,
			 GummyPath,
			 GummyIdent,
			 RealmName)
			VALUES (?,?,?,?,?)", {},

		 $param->{'GUMMY'},
		 $param->{'CLASS'},
		 $param->{'GPATH'},
		 $param->{'GROUP'},
		 $realm);


		# Audit action
		$audit->GummyAdd($realm,$param->{'GUMMY'});

		# Cleanup
		$param->Cleanup('GUMMY','GROUP','GPATH','Add');
	}



	# Query management
	# Prepare and execute
	$statement = $data->prepare("SELECT 
		 GummyName,
		 GummyPath,
		 GummyIdent
		FROM  Gummy
		WHERE GummyClass=?
		  AND RealmName=?
		ORDER BY GummyIdent, GummyName");
	$statement->execute($param->{'CLASS'}, $realm);


	# Form handle
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat());

	# HTML display
	# Table Top
	print "<CENTER>\n<TABLE WIDTH=500 CELLPADDING=5>\n";

	# Cycle through gummies
	while ($res = $statement->fetchrow_hashref())
	{
		# Padding
		$ident = (defined($res->{'GummyIdent'})) ? $res->{'GummyIdent'} : 'None';

		# Title
		if ($ident ne $group)
		{
			# Alternate
			$group = $ident;

			# Uppercase
			$ident = uc($ident);


			# Display row
			print " <TR>\n",
			      "  <TD COLSPAN=3><BR>\n<BR>\n<B><FONT SIZE=+2>$ident</FONT></B><BR></TD>\n",
			      " </TR>\n";
		}

		# Name handling
		$name = $res->{GummyName};
		$name =~ s/\s//gs;

		# Handle removal
		if (defined($param->{$name}))
		{
			# Remove gummy
			$data->do("DELETE FROM Gummy
				WHERE GummyName=?
				  AND RealmName=?", {},

			 $res->{'GummyName'},
			 $realm);


			# Audit action
			$audit->GummyRem($res->{'GummyName'},$realm);

			# Cleanup
			$param->Cleanup($name);
		}

		# Display if not removed
		else
		{
			# Display row
			print " <TR>\n",
			      "  <TD><IMG SRC=\"$res->{GummyPath}\"></TD>\n",
			      "  <TD><B>$res->{GummyName}</B></TD>\n",
			      "  <TD><CENTER>", $cgi->checkbox(-name=>$name, -label=>''), "</CENTER></TD>\n",
			      " </TR>\n";
		}
	}

	# Display row
	print " <TR>\n",
	      "  <TD>&nbsp;</TD>\n",
	      "  <TD>&nbsp;</TD>\n",
	      "  <TD><CENTER>", $cgi->submit('Remove'), "</CENTER></TD>\n",
	      " </TR>\n";


	# End of table
	print "</TABLE>\n</CENTER>\n";

	# End form
	print $cgi->end_form(), "\n";
	
	# End query
	$statement->finish();


	# Spacer
	print "<BR>\n<BR>\n";

	# Form handle
	print $cgi->start_form();
	print $param->EmbedNormal('USER', $param->{'USER'}, 'CRYPT',  $param->{'CRYPT'}, 'CLASS',  $param->{'CLASS'});

	# Addition section
	# Table top
	$table->MakeTop();
	$table->MakeSingle("<B><FONT SIZE=+2>ADDITION</FONT></B>");
	$table->MakeBlank();

	# Display rows
	$table->MakeValid('<B>Name</B>',  $cgi->textfield('GUMMY', '', 20, 20));
	$table->MakeValid('<B>Group</B>', $cgi->textfield('GROUP', '', 20, 20));
	$table->MakeBlank();

	$table->MakeSingle('<B>Full Path</B>');
	$table->MakeSingle($cgi->textfield('GPATH', 'http://', 75, 200));
	$table->MakeBlank();

	$table->MakeSingle($cgi->submit('Add'));
	$table->MakeBottom();
	
	
	# End form
	print $cgi->end_form(), "\n";
}


#####################
# Gummy Selection
#
# Very simple sub-routine that will simply allow the user to select the class
# of gummies to search for.

sub GummySelect
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter reference
	my $table    = shift;					# Table handler
	my $widget;						# Displayed widget

	my @class;						# Table array


	#####################
	# Program area

	# Gummy list 
	@class  = qw(admin default canon contributor general mod super);

	# Generate widget
	$widget = $cgi->popup_menu('CLASS', \@class, $class[0]);


	# Form handle
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat());

	# Table display
	# Top and widget
	$table->MakeTop();
	$table->MakeValid('<B>Select Class</B>', $widget);
	$table->MakeBlank();

	# Submit and bottom
	$table->MakeSingle($cgi->submit('Select'));
	$table->MakeBottom();

	# End form
	print $cgi->end_form();
}


#####################
# Realm Addition
#
# Sub-routine that displays additions of new realms.

sub RealmAddition
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter list
	
	my $table    = new Ethereal::Table($cgi);		# Tale handle

	my @radiolang   = qw(relaxed restricted);		# Language
	my @radioaccess = qw(relaxed restricted list);		# Access
	my @radioguest  = qw(yes no post);			# Guests
	my @radiotags   = qw(relaxed restricted);		# Tags
	my @radiotype	= qw(yes no hybrid alias);		# Realm type

	# Initialize radio buttons
	my $radiolang   = $cgi->radio_group('NEWLANG',  \@radiolang,  $radiolang[0],  'true');
	my $radioaccess = $cgi->radio_group('NEWACCESS',\@radioaccess,$radioaccess[0],'true');
	my $radioguest  = $cgi->radio_group('NEWGUEST', \@radioguest, $radioguest[0], 'true');
	my $radiotags   = $cgi->radio_group('NEWTAGS',  \@radiotags,  $radiotags[0],  'true');
	my $radiotype	= $cgi->radio_group('NEWTYPE',  \@radiotype,  $radiotype[0],  'true');

	
	#####################
	# Program area

	# Cleanup
	$param->Cleanup('ADDITION');


	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";

	# Prompt and display realm name
	$table->MakeTop();
	$table->MakeValid('<B>Realm Name</B>',   $cgi->textfield('NEWNAME'  , '', 20, 30));
	$table->MakeValid('<B>Aliased Realm</B>',   $cgi->textfield('NEWALIAS'  , '', 20, 30));
	$table->MakeValid('<B>Realm Type</B>',   $radiotype);
	$table->MakeBlank();

	# Image restrictions
	$table->MakeValid('<B>Image Height</B>', $cgi->textfield('NEWHEIGHT', '0', 20, 5));
	$table->MakeValid('<B>Image Width</B>',  $cgi->textfield('NEWWIDTH' , '0', 20, 5));
	$table->MakeValid('<B>Image Size</B>',  $cgi->textfield('NEWSIZE' , '0', 20, 3));
	$table->MakeBlank();

	# Language selection
	$table->MakeValid('<B>Language</B>',        $radiolang);
	$table->MakeBlank();

	# Access type
	$table->MakeValid('<B>Access type</B>',     $radioaccess);
	$table->MakeBlank();

	# Guest access
	$table->MakeValid('<B>Guest Access</B>',    $radioguest);
	$table->MakeBlank();

	# Tag enforcement
	$table->MakeValid('<B>Tag enforcement</B>', $radiotags);
	$table->MakeBlank();

	# Submit and reset
	$table->MakeValid($cgi->submit('Add'), $cgi->reset('Clear'));
	$table->MakeBottom();

	# End form
	print $cgi->end_form(), "\n";


}

#####################
# RealmDisplay
#
# Sub-routine that will display and handle basic operations reguarding direction
# of calls.   As in remove and additions.

sub RealmDisplay
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter list reference

	my $data     = $database->{'HANDLE'};			# Database connection handle
	my $statement; 						# Query handle

	my $audit;						# Audit handle 
	my $date     = new Ethereal::Date();			# Date handler

	my $count    = 0;					# Colour changer
	my $naughty  = 'system';				# Naughty word
	my $url      = $cgi->url(-absolute=>1);			# Self referencing URL

	my $in_back;						# Back button inline
	my $in_add;						# Inline button inline
	my $handle;						# List of supervisory handles
	my $values;						# Received values

	my $realm;						# Realm name modified
	my $status;						# Realm status

	my @super;						# List of supervisors


	#####################
	# Program area

	# Audit handle
	$audit = new Ethereal::Audit($database, $cgi, $param);

	# Data addition
	if (defined( $param->{'NEWNAME'} ))
	{
		# New aliases
		$param->{'NEWALIAS'} = (length($param->{'NEWALIAS'}) > 2) ? $param->{'NEWALIAS'} : undef;

		# Execute
		$data->do("INSERT INTO Realm 
			(RealmName,
			 RealmAlias,
			 RealmLanguage,
			 RealmAccess,
			 RealmGuests,
			 RealmPublic,
			 RealmTags,
			 RealmImageHeight,
			 RealmImageWidth,
			 RealmImageSize,
			 RealmTimestamp)
			VALUES (?,?,?,?,?,?,?,?,?,?,?)", {}, 

		 $param->{'NEWNAME'}, 
		 $param->{'NEWALIAS'}, 
		 $param->{'NEWLANG'},
		 $param->{'NEWACCESS'}, 
		 $param->{'NEWGUEST'},
		 $param->{'NEWTYPE'},
		 $param->{'NEWTAGS'}, 
		 $param->{'NEWHEIGHT'},
		 $param->{'NEWWIDTH'}, 
		 $param->{'NEWSIZE'},
		 time);

		# Audit Trail
		$audit->RealmAddition($param->{'NEWNAME'});

		# Cleanup
		$param->Cleanup('NEWNAME');
	}

	# Links
	# Generate inline
	($in_back) = $param->EmbedInline('USER', $param->{'USER'}, 'CRYPT', $param->{'CRYPT'});
	($in_add)  = $param->EmbedInline($param->Flat(), 'ADDITION', 'True');

	print "<CENTER><P>",
	      "<A HREF=\"$url$in_back\">[Back]</A> ",
	      "<A HREF=\"$url$in_add\">[Addition]</A>",
	      "</P</CENTER>";

	# Data viewing
	# Prepare and execute
	$statement = $data->prepare("SELECT RealmName, 
		 RealmAlias, 
		 RealmLanguage, 
		 RealmAccess, 
		 RealmGuests, 
		 RealmTags, 
		 RealmTimestamp 
		FROM  Realm 
		WHERE RealmPublic=? 
		ORDER BY RealmName");
	$statement->execute($param->{'REALMTYPE'});


	# Start form
	print $cgi->start_form();


	# Table name
	print "<TABLE WIDTH=\"100%\">";

	# Display header
	print " <TR>\n";
	print "  <TD><B>Realm Name</B></TD>\n";

	# Non aliased
	if ($param->{'REALMTYPE'} ne 'alias') 
	{
		print "  <TD WIDTH=\"50\"><B>Language</B></TD>\n";
		print "  <TD WIDTH=\"50\"><B>Access</B></TD>\n";
		print "  <TD WIDTH=\"50\"><B>Guests</B></TD>\n";
		print "  <TD WIDTH=\"50\"><B>Tags</B></TD>\n";
		print "  <TD WIDTH=\"50\"><B>Creation</B></TD>\n";
	}

	# Aliased
	else
	{
		print "  <TD WIDTH=\"50\"><B>Alias</B></TD>\n";
	}


	# Fetch rows of data
	# Handle specific realms
	while ($values = $statement->fetchrow_hashref())
	{
		# Realm name
		$realm = $values->{'RealmName'};
                $realm = uc($realm);
		$realm =~ s/\W//gs;


		# Handling of changes
		if (defined($param->{$realm}))
		{
			# Upgrade or downgrade a realm
			if ((defined($param->{'Promote'})) || (defined($param->{'Demote'})))
			{
				# Upgrades
				if (defined($param->{'Promote'}))
				{
					# Only hybrids and private realms can be upgraded
					if ($param->{'REALMTYPE'} eq 'hybrid') { $status = 'yes'; }
					elsif ($param->{'REALMTYPE'} eq 'no')  { $status = 'hybrid'; };
				}

				# Downgrades
				else
				{
					# Only public can be brought down
					if ($param->{'REALMTYPE'} eq 'yes')    { $status = 'hybrid'; }
				}


				# Commit changes
				if (defined($status))
				{
					# Database update
					$data->do("UPDATE Realm
						SET RealmPublic=?
						WHERE RealmName=?", {},
					 $status,
					 $values->{'RealmName'});


					# Audit entry
					$audit->RealmStatus($values->{'RealmName'}, $status);
	

					# Undefine values
					$status = undef;
					$realm  = undef;
				}

				# Cleanup
				$param->Cleanup($realm);
			}


			# Handling of removal
			if (defined($param->{'Remove'}))
			{
				# Remove all dangling supervisors/moderators
				$database->GetList(\@super, "SELECT p.PuppeteerLogin
					FROM  Puppeteer p, RealmSupervisor r
					WHERE p.PuppeteerLogin = r.PuppeteerLogin
					AND r.RealmName=?
					ORDER BY p.PuppeteerLogin", $values->{'RealmName'});

				# Cycle though
				foreach $handle (@super)
				{
					# Determin amount of instances
					my ($val) = $database->DataGet("SELECT COUNT(*) 
						FROM RealmSupervisor
						WHERE PuppeteerLogin=?", $handle);

					# Remove priviliges to all those that exist once
					if ((defined($val)) && ($val == 1))
					{
						$data->do("UPDATE Puppeteer 
							SET   PuppeteerSuperPrivs='no',
							      PuppeteerModPrivs='no'
							WHERE PuppeteerLogin=?", {}, $handle);
					} 		
				}


				# Source handles
				$data->do("DELETE FROM Puppet 
					 WHERE PuppetPool=? 
					   AND PuppetPool <> ?", {}, 

				 $values->{'RealmName'}, 
				 $naughty);

				# Gummies
				$data->do("DELETE FROM Gummy
					 WHERE RealmName=? 
					   AND RealmName <> ?", {}, 

				 $values->{'RealmName'}, 
				 $naughty);


				# Remove from satellite tables
				$data->do("DELETE FROM Board WHERE RealmName=?", {}, $values->{'RealmName'});
				$data->do("DELETE FROM BoardTopic WHERE RealmName=?", {}, $values->{'RealmName'});
				$data->do("DELETE FROM PuppeteerRestriction WHERE RealmName=?", {}, $values->{'RealmName'});
				$data->do("DELETE FROM PuppetInvitation WHERE RealmName=?", {}, $values->{'RealmName'});
				$data->do("DELETE FROM RealmSupervisor WHERE RealmName=?", {}, $values->{'RealmName'});
				$data->do("DELETE FROM Restriction WHERE RealmName=?", {}, $values->{'RealmName'});
				$data->do("DELETE FROM RealmDesign WHERE RealmName=?", {}, $values->{'RealmName'});
				$data->do("DELETE FROM RealmLocale WHERE RealmName=?", {}, $values->{'RealmName'});

				# Remove aliases
				$data->do("DELETE FROM Realm WHERE RealmAlias=?", {}, $values->{'RealmName'});

				# Delete parent table
				$data->do("DELETE FROM Realm WHERE RealmName=?", {}, $values->{'RealmName'});

				# Audit Trail
				$audit->RealmRemoval($values->{'RealmName'});


				# Removal cleanup
				$param->Cleanup($realm);

				# Undefine realm
				$realm = undef;
			}
		}


		# Only process if there is a need
		if (defined($realm))
		{
			# Colour handling
			# Determin colour of row
			if (($count % 2) != 0) 
			{ print "\t<TR>\n"; }
			else
			{ print "\t<TR BGCOLOR=\"\#999999\">\n"; }
		
			# Increment	
			$count++;
	


			# Realm name
			print "  <TD><B>$values->{RealmName}</B></TD>\n";

			# Non aliased
			if ($param->{'REALMTYPE'} ne 'alias') 
			{
				print "  <TD><I>$values->{RealmLanguage}</I></TD>\n";
				print "  <TD><I>$values->{RealmAccess}</I></TD>\n";
				print "  <TD><I>$values->{RealmGuests}</I></TD>\n";
				print "  <TD><I>$values->{RealmTags}</I></TD>\n";
				print "  <TD><I>", $date->GetDate($values->{RealmTimestamp}), "</I></TD>\n";
			}

			# Aliased
			else
			{
				print "  <TD><I>$values->{'RealmAlias'}</I></TD>\n";
			}


			# Realm handling
			print  "  <TD>", $cgi->checkbox(-name=>$realm, -label=>''), "</TD>";
		}
	}

	# End query
	$statement->finish();

	# Close table
	print "</TABLE>\n";


	# Final form elements
	if ($param->{'REALMTYPE'} ne 'alias')
	{
		# Allow up upgrades and downgrades
		print "<BR><CENTER>",
		 $cgi->submit('Remove'), "&nbsp;&nbsp;&nbsp;",
		 $cgi->submit('Promote'), "&nbsp;",
		 $cgi->submit('Demote'), "&nbsp;",
		 "</CENTER>";
	}
	else
	{
		# Only allow for removal
		print "<BR><CENTER>",
		 $cgi->submit('Remove'),
		 "</CENTER>";
	}


	# Cleanup
	$param->Cleanup(
	 'Remove',
	 'Promote',
	 'Demote'
	);


	# End form
	print $param->EmbedNormal($param->Flat()), "\n";
	print $cgi->end_form(), "\n";
}


#####################
# Realm Select
#
# Basic display and prompt user for a selection on which realm to adminster and
# make modifications of some type to.

sub RealmSelect
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash

	my $table    = new Ethereal::Table($cgi);		# Table handler

	my @realms;						# Array of realms
	my $realms;						# Popup menu

	
	#####################
	# Program area

	# Prepare of drop down
	$database->GetListNonPrivate(\@realms);
	$realms = $cgi->popup_menu('ROOM', \@realms);

	# HTML Display
	# Star form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";

	# Start table
	$table->MakeTop();

	# Realm selection
	$table->MakeValid('<B>Select Realm</B>', $realms);
	$table->MakeBlank();

	# Submit and bottom
	$table->MakeSingle($cgi->submit('Continue'));
	$table->MakeBottom();

	# End of form and embed
	print $cgi->end_form(), "\n"; 
}


#####################
# Realm Type
#
# Simply allows the user to determine which type of rooms to view private or 
# public based on a simple statement.

sub RealmType						# Database, CGI
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter list

	my $table    = new Ethereal::Table($cgi);		# Table handler

	my $list;	
	my @list     = qw(yes no hybrid alias);			# List of choices


	#####################
	# Program area
		
	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";

	# Prepare list
	$list = $cgi->radio_group('REALMTYPE', \@list, $list[0], 'true');

	# Prompt and Display Document choices
	$table->MakeTop();
	$table->MakeValid('<B>Public Realm</B>', $list);
	$table->MakeBlank();

	$table->MakeValid($cgi->submit('List Realms'), ' ');
	$table->MakeBottom();

	# End form
	print $cgi->end_form(), "\n";
}


#####################
# SQL Handler
# 
# Looks at parameters and returns a formatted SQL statement to be used for searches

sub SqlHandler							# Parameter Hash
{
	#####################
	# Data members
	my $param    = shift;					# Parameter hash
	my $sql;						# SQL query	

	my @range = qw(Administrators Supervisors Banned Verified Non-Priviledged All-Encompasing);
	my @field = qw(E-Mail Login Name Puppet Homepage);


	#####################
	# Program area

	# Initial construction of mun
	$sql = "SELECT mun.PuppeteerLogin, mun.PuppeteerName, mun.PuppeteerEMail";
	$sql = "$sql FROM Puppeteer mun LEFT JOIN Puppet pup ON mun.PuppeteerLogin=pup.PuppeteerLogin";


	# Build SQL query
	# Admin privs
	if ($param->{'RANGE'} eq $range[0])
	{ $sql = "$sql WHERE mun.PuppeteerAdminPrivs='yes'"; }


	# Supervisor privs
	elsif ($param->{'RANGE'} eq $range[1])
	{ $sql = "$sql WHERE mun.PuppeteerSuperPrivs='yes'"; }
	
	# Banned users
	elsif ($param->{'RANGE'} eq $range[2])
	{ $sql = "$sql WHERE mun.PuppeteerBan='yes' AND mun.PuppeteerAlias IS NULL"; }

	# Verified users
	elsif ($param->{'RANGE'} eq $range[3])
	{ $sql = "$sql WHERE mun.PuppeteerVerified='yes'"; }
	
	# Non-Priviledged users
	elsif ($param->{'RANGE'} eq $range[4])
	{ $sql = "$sql WHERE mun.PuppeteerSuperPrivs='no' AND mun.PuppeteerAdminPrivs='no'"; }

	# Only complete following if nothing (or near to) in keywords
	unless (length($param->{'KEYWORD'}) < 3)
	{
		# Verification on data shaping
		if ($sql =~ /WHERE/)
		{ $sql = "$sql AND"; }
		else
		{ $sql = "$sql WHERE"; }
		

		# Field selection
		# E-Mail field search
		if ($param->{'FIELD'} eq $field[0])
		{ $sql = "$sql mun.PuppeteerEMail"; }

		# Login name search
		elsif ($param->{'FIELD'} eq $field[1])
		{ $sql = "$sql mun.PuppeteerLogin"; }

		# Full name search
		elsif ($param->{'FIELD'} eq $field[2])
		{ $sql = "$sql mun.PuppeteerName"; }

		# Puppet name search
		elsif ($param->{'FIELD'} eq $field[3])
		{ $sql = "$sql pup.PuppetName"; }

		# Homepage name search
		elsif ($param->{'FIELD'} eq $field[4])
		{ $sql = "$sql mun.PuppeteerHome"; }

		# Completion of SQL statement
		$sql = "$sql LIKE '\%$param->{'KEYWORD'}\%'";
	}

	# Return statement to parent
	return "$sql GROUP BY PuppeteerLogin ORDER BY PuppeteerLogin";
}


#####################
# User Lookup
#
# Will perform a check against the IP addresses that the user has.   Based
# on that information will come up with other matches.

sub UserLookup
{
	#####################
	# Data members

	my $database = shift;					# Database handle
	my $param    = shift;					# Parameter reference

	my $data     = $database->{'HANDLE'};			# Direct database handle
	my $statement;						# Database statement

	my $date     = new Ethereal::Date();			# Date handle
	my $send;						# File send
	my $tstamp   = $date->GetDateShort(time);		# Timestamp

	my $count;						# Misc. counter
	my $email;						# Email address
	my $host;						# Singular hostname
	my $match;						# Fount match
	my $placeholders;					# Number of servers to conduct

	my @aliases;						# Known aliases 
	my @hosts;						# List of hosts

	my %ranges;						# Address ranges
	my %system;						# System variables

	#####################
	# Program Area

# Format for the output
my $LISTING = '
     @<<<<<<<<<<        @>>>>>>>>>>        @>>>>>>>>>>';

my $HEADER = '

   =======================================================
    Range :  @<<<<<<<<<<<<<<
    ----------------------------------------------------
    (Matches )         (Occurrences)      (Last used )';


	# Link with hash
	$database->GetHashSystem(\%system);

	# Retreive Email
	($email) = $database->DataGetPuppeteerEmail($param->{'USER'}); 


	# Get list
	$database->GetList(\@aliases, "SELECT PuppeteerLogin
		FROM Puppeteer
		WHERE PuppeteerAlias=?
		ORDER BY PuppeteerLogin", $param->{'LOGIN'});

	# Append login
	unshift(@aliases, $param->{'LOGIN'});

	# Aliases
	$count = @aliases;


	# Generate placeholders
	$placeholders = "?," x $count;
	chop($placeholders);


	# Hostnames
	$database->GetList(\@hosts, "SELECT PuppeteerHostname
		FROM  Audit
		WHERE PuppeteerLogin IN ($placeholders)
		GROUP BY PuppeteerHostname", @aliases);


	# Discover ranges
	while ($host = shift(@hosts))
	{
		# Process if this is an IP address
		if ($host =~ /(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})/)
		{
			# Determin range
			$host = "$1.$2.$3";

			# Populate hash
			$ranges{$host} = defined($ranges{$host}) ? 1 : $ranges{$host}++;	
		}
	}	


	# Send mail
	# Create instance
	$send = tie(*MAIL, 'Ethereal::Mail', $database);

	# Initial setup
	$send->SetSubject("$param->{LOGIN} cross-reference ($tstamp)");

	# Recipients
	$send->AddTo($email);


	# Cycle through range
	foreach $host (sort(keys(%ranges)))
	{
		# Search through ranges
 		$statement = $data->prepare("SELECT PuppeteerLogin, COUNT(*), MAX(AuditTimestamp)
			FROM  Audit
			WHERE PuppeteerHostname LIKE '${host}%' 
			GROUP BY PuppeteerLogin
			ORDER BY PuppeteerLogin");
		$statement->execute();


		# Header
		$send->Write($HEADER, $host);


		# Display full list
		while (($match, $count, $tstamp) = $statement->fetchrow())
		{
			$send->Write($LISTING, $match, $count, $tstamp);
		}

		# Finish query
		$statement->finish();
	}

	# Close and send
	close(MAIL);
}


#####################
# User Modify
#
# Will either prompt for information or make changes and list the changed 
# values to the user.   Only one value can be changed at a time.  

sub UserModify							# Database, CGI, Parameter reference
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter reference

	my $audit;						# Audit handle
	my $date;						# Date handle
	my $table;						# Table handle
	my $page;						# Page handle

	my $data     = $database->{'HANDLE'};			# Database connection handle			
	my $statement;						# Statement handle

	my $radioadmin;						# Radio adminstration
	my $radioban;						# Radio ban
	my $radiomod;						# Radio moderator
	my $radiosuper;						# Radio supervisory 
	my $radiover;						# Radio verivified user

	my $alias;						# Singular alias
	my $aliases;						# List of aliases

	my $hostname;						# Individual hostname
	my $hostnames= '';					# List of hostnames

	my $puppet;						# Individual puppet
	my $puppets  = '';					# List of puppets

	my $home;						# Homepage directory
	my $inline;						# Inline parameters
	my $login;						# Login name
	my $salt;						# Password salt
	my $url      = $cgi->url(-absolute=>1);			# Return URL
	my $values;						# Hash reference

	my $pool;						# Puppet pool
	my $puppet;						# Singular puppet
	my $reg;						# Registration flag

	my $ltime;						# Nicely formatted date
	my $htime;						# Nicely formatted date

	my @aliases;						# List of aliases
	my @choices = qw(yes no);				# Yes or no
	my @hostnames;						# Hostnames
	my @puppets;						# List of puppets

	my %home;						# Homepage handler


	#####################
	# Program area

	# Audit handle
	$audit = new Ethereal::Audit($database, $cgi, $param);
	$date  = new Ethereal::Date();
	$table = new Ethereal::Table($cgi);

	# Change username to normal
	$login = $param->{'LOGIN'};

	# Home Page Handler
	if (defined($param->{'HOME'}))
	{
		#####################
		# Data Members
		my $year = '31536000';				# Physical year
		my $roof = '2147483647';			# Maximum size for time
		my $time = time;				# Current time


		# Define time
		$htime = (defined($param->{'POS'})) ? $param->{'POS'} : $param->{'NEG'};
		$htime = ($htime =~ /^\d*/) ? $htime : $time;
		$htime = (($htime > $time) && ($htime < $roof)) ? $htime : $time;			

		# Final calculation
		$htime = (defined($param->{'POS'})) ? ($htime+$year) : ($htime-$year);

		# Database update
		$data->do("UPDATE Puppeteer
			SET   PuppeteerHomeSpan=?
			WHERE PuppeteerLogin=?", {},
			$htime,
			$login);
		
		# Cleanup
		$param->Cleanup(
			'HOME',
			'POS',
			'NEG'
		 );
	}

	# Cross reference handler
	if (defined($param->{'REFERENCE'}))
	{
		# Call cross reference
		UserLookup($database, $param);	
	}


	# If save then save
	if (defined($param->{'SAVE'}))				# Saved parameters
	{
		# Password reset
		if ($param->{'PASSWORD'} =~ /^\*\w/)
		{
			# Determine password
			$salt = md5_hex(rand());
			$salt = substr($salt, 0, 10);

			# Strip star
			$param->{'PASSWORD'} =~ s/^\*//;
			
			# Generate crypt
			$param->{'PASSWORD'} = crypt($param->{'PASSWORD'}, $salt);


			# Audit entry
			$audit->PasswordChange($login);
		}

		# Handling of ban
		if ((defined($param->{'BANCONF'})) && ($param->{'BAN'} eq 'yes'))
		{	
			# Home page handling
			# Link to hash
			$database->GetHashHome(\%home);

			# Create handle
			$page   = new Ethereal::Home($database);

			# Retreive home
			($home) = $database->DataGetPuppeteerHome($login);

			# Remove folder
			if (defined($home)) 
			{ 
				$page->DirClear("$home{SetRoot}/$home"); 
			}	


			# Handle moving
			# Recover puppets
			$statement = $data->prepare("SELECT 
				 PuppetName,
				 PuppetPool
				FROM Puppet
				WHERE PuppeteerLogin=?");
			$statement->execute($login);


			# Cycle through
			while (($puppet, $pool) = $statement->fetchrow())
			{
				# System handles
				if ($pool eq 'system')
				{
					# Change classification
					$data->do("UPDATE Puppet
						SET   PuppetPool=?
						WHERE PuppetName=?
						  AND PuppeteerLogin=?", {},
					
					 'banned',
					 $puppet,
					 $login);
				}

				# Source handles
				elsif (($pool ne 'system') 
				   && ($pool ne 'banned')) 
				{
					# Change classification
					$data->do("DELETE FROM Puppet
						WHERE PuppetName=?
						  AND PuppeteerLogin=?", {},
					
					 $puppet,
					 $login);
				}
			}		

			# End statement
			$statement->finish();


			# Determine alias
			$param->{'ALIAS'} = (length($param->{'ALIAS'}) > 2) ? $param->{'ALIAS'} : undef;


			# Set attribute
			$data->do("UPDATE Puppeteer
				SET PuppeteerBan='yes',
				    PuppeteerAlias=?
				WHERE PuppeteerLogin=?", {},

			 $param->{'ALIAS'},
			 $login);


			# Remove any unmoved puppets
			$data->do("DELETE FROM Puppet
				WHERE PuppeteerLogin=?
				  AND PuppetPool=?", {},

			 $login,
			 'system');


			# Star out password
			$param->{'PASSWORD'} = '*************';


			# Audit Trail
			$audit->BanSystem($login);
		}	


		# Make changes
		$data->do("UPDATE Puppeteer 
			 SET PuppeteerPassword=?,  
			     PuppeteerName=?,
			     PuppeteerEMail=?, 
			     PuppeteerAdminPrivs=?, 
			     PuppeteerSuperPrivs=?,
			     PuppeteerModPrivs=?,
			     PuppeteerVerified=?
		 	WHERE PuppeteerLogin=?",  {}, 
	 	 $param->{'PASSWORD'}, 
		 $param->{'FULLNAME'}, 
 		 $param->{'EMAIL'}, 
		 $param->{'ADMIN'}, 
		 $param->{'SUPER'},
		 $param->{'MOD'}, 
		 $param->{'VER'},
		 $login );

		# Audit trail
		$audit->PuppeteerChange($login);

		# Cleanup
		$param->Cleanup('SAVE', 'BANCONF');
	}


	# Back button
	# Generate inline
	($inline) = $param->EmbedInline('USER', $param->{'USER'}, 'CRYPT', $param->{'CRYPT'}, 'RANGE', $param->{'RANGE'}, 'KEYWORD', $param->{'KEYWORD'}, 'FIELD', $param->{'FIELD'}, 'SEARCH', $param->{'SEARCH'});

	# Display
	print "<CENTER><A HREF=\"$url$inline\">[Back]</A></CENTER>";	


	# Data display
	# Prepare and execute
	$statement = $data->prepare("SELECT PuppeteerLogin,
		    PuppeteerPassword,
		    PuppeteerName,
		    PuppeteerAlias,
		    PuppeteerEMail,
		    PuppeteerTimestamp,
		    PuppeteerFormatting,
		    PuppeteerChatPref,
		    PuppeteerJavascript,
		    PuppeteerAdminPrivs,
		    PuppeteerSuperPrivs,
		    PuppeteerModPrivs,
		    PuppeteerBan,
		    PuppeteerHostname,
		    PuppeteerHome,
		    PuppeteerHomeSpan,
		    PuppeteerVerified
		FROM  Puppeteer 
		WHERE PuppeteerLogin=?");
	$statement->execute($login);

	# Retreive and finish
	$values = $statement->fetchrow_hashref();
	$statement->finish();


	# Registration flag
	($reg) = $database->DataGetPuppeteerReg($login);


	# Recover Aliases
	if ($values->{'PuppeteerBan'} eq 'yes')
	{
		# Get list
		$database->GetList(\@aliases, "SELECT PuppeteerLogin
			FROM Puppeteer
			WHERE PuppeteerAlias=?
			ORDER BY PuppeteerLogin", $login);

		# Append login name
		unshift(@aliases, $login);

		# Join into singular line
		$aliases = join("\n", @aliases);


		# Recover Puppets
		foreach $alias (@aliases)
		{
			# Recover puppets
			$database->GetList(\@puppets, "SELECT PuppetName 
				FROM Puppet 
				WHERE PuppeteerLogin=?
				ORDER BY PuppetName", $alias);

			# Title
			$puppets = "$puppets$alias:\n-----------\n";

			# Construct alias 
			while ($puppet = shift(@puppets))
			{
				# Add all hostnames
				$puppets = "$puppets\t$puppet\n";			
			}

			# Spacer 
			$puppets = "$puppets\n\n";			


			# Recover IPs
			$database->GetList(\@hostnames, "SELECT PuppeteerHostname 
				FROM  Audit 
				WHERE PuppeteerLogin=? 
				GROUP BY PuppeteerHostname 
				ORDER BY PuppeteerHostname", $alias);

			# Title
			$hostnames = "$hostnames$alias:\n-----------\n";

			# Construct alias 
			while ($hostname = shift(@hostnames))
			{
				# Add all hostnames
				$hostnames = "$hostnames\t$hostname\n";			
			}

			# Spacer 
			$hostnames = "$hostnames\n\n";			
		}	
	}

	# Standard handling
	else
	{
		# Recover IPs
		$database->GetList(\@hostnames, "SELECT PuppeteerHostname 
			FROM  Audit 
			WHERE PuppeteerLogin=? 
			GROUP BY PuppeteerHostname 
			ORDER BY PuppeteerHostname", $login);

		# Recover puppets
		$database->GetList(\@puppets, "SELECT PuppetName 
			FROM Puppet 
			WHERE PuppeteerLogin=?
			ORDER BY PuppetName", $login);
	
		# Join into singular line
		$hostnames = join("\n", @hostnames);
		$puppets   = join("\n", @puppets);
	}


	# Adapt timestamp
	$ltime = $date->GetDate($values->{'PuppeteerTimestamp'});
	$htime = $date->GetDate($values->{'PuppeteerHomeSpan'});


	# Start HTML
	# Start form
	print $cgi->start_form(), "\n";
	print $param->EmbedNormal($param->Flat(), 'SAVE', 'True'), "\n";


	# Table top
	$table->MakeTop();

	# Registration flag
	if (defined($reg))
	{
		# Invalid registraion notice
		$table->MakeSingle('<CENTER><B>REGISTRATION INCOMPLETE</B></CENTER>');
		$table->MakeBlank();
	}


	# Login, Password
	$table->MakeValid('<B>Login Name</B>', "<B>$login</B>");
	$table->MakeValid('<B>Password</B>',  $cgi->textfield('PASSWORD', $values->{'PuppeteerPassword'}, 25, 13));
	$table->MakeBlank();

	# Name, Email, Alias
	$table->MakeValid('<B>Full Name</B>', $cgi->textfield('FULLNAME', $values->{'PuppeteerName'}, 25, 30));
	$table->MakeValid('<B>E-Mail</B>',    $cgi->textfield('EMAIL', $values->{'PuppeteerEMail'}, 25, 45));
	$table->MakeValid('<B>Alias</B>',     $cgi->textfield('ALIAS', $values->{'PuppeteerAlias'}, 25, 10));
	$table->MakeBlank();

	# Timestamp, Hostname (cannot be changed)
	$table->MakeValid('<B>Last login</B>', "<I>$ltime</I>");
	$table->MakeValid('<B>Hostname</B>', "<I>$values->{'PuppeteerHostname'}</I>");
	$table->MakeBlank();


	# Generate inline
	($inline) = $param->EmbedInline($param->Flat(), 'HOME', 'True');

	# Cross reference
	$table->MakeValid('<B>Cross Reference</B>', "<A HREF=\"$url$inline&REFERENCE=True\">Check</A>");
	$table->MakeBlank();

	# Homepage Information
	$table->MakeValid('<B>Home Page</B>', "<I>$values->{'PuppeteerHome'}</I>");
	$table->MakeValid('<B>Expires</B>', "<I>$htime</I>");
	$table->MakeValid('&nbsp;', "<A HREF=\"$url$inline&POS=$values->{PuppeteerHomeSpan}\">Add Year</A>");
	$table->MakeValid('&nbsp;', "<A HREF=\"$url$inline&NEG=$values->{PuppeteerHomeSpan}\">Remove Year</A>");
	$table->MakeBlank();

	# Formatting, Chat Prefs and Javascript (cannot be changed)
	$table->MakeValid('<B>Formatting</B>', "<I>$values->{'PuppeteerFormatting'}</I>");
	$table->MakeValid('<B>Chat Prefs</B>', "<I>$values->{'PuppeteerChatPref'}</I>");
	$table->MakeValid('<B>Javascript</B>', "<I>$values->{'PuppeteerJavascript'}</I>");
	$table->MakeBlank();

	# Preperation of radio button labels
	$radioadmin = $cgi->radio_group('ADMIN', \@choices, $values->{'PuppeteerAdminPrivs'});
	$radioban   = $cgi->radio_group('BAN',   \@choices, $values->{'PuppeteerBan'});
	$radiosuper = $cgi->radio_group('SUPER', \@choices, $values->{'PuppeteerSuperPrivs'});
	$radiomod   = $cgi->radio_group('MOD'  , \@choices, $values->{'PuppeteerModPrivs'});
	$radiover   = $cgi->radio_group('VER'  , \@choices, $values->{'PuppeteerVerified'});

	# Adminstrative and Supervisory controls
	$table->MakeValid('<B>Administrative</B>', $radioadmin);
	$table->MakeValid('<B>Supervisory   </B>', $radiosuper);
	$table->MakeValid('<B>Moderator     </B>', $radiomod);
	$table->MakeValid('<B>Verified      </B>', $radiover);
	$table->MakeBlank();

	# User banning
	$table->MakeValid('<B>Banned        </B>', $radioban);
	$table->MakeValid('&nbsp;', $cgi->checkbox(-name=>'BANCONF', -label=>'Confirm'));
	$table->MakeBlank();

	# List of aliases
	if ($values->{'PuppeteerBan'} eq 'yes')
	{
		$table->MakeSingle('<B>Known Aliases</B>');
		$table->MakeSingle($cgi->textarea('ALIASES', $aliases, 10, 40));
		$table->MakeBlank();
	}

	# List of Puppets
	$table->MakeSingle('<B>Puppets</B>');
	$table->MakeSingle($cgi->textarea('PUPPETS', $puppets, 10, 40));
	$table->MakeBlank();

	# IP History
	$table->MakeSingle('<B>Hostname History</B>');
	$table->MakeSingle($cgi->textarea('HOSTNAMES', $hostnames, 10, 40));
	$table->MakeBlank();

	# Submit and Reset
	$table->MakeValid($cgi->submit('Change'), $cgi->reset('Reset'));
	$table->MakeBottom();

	# End form 
	print $cgi->end_form(), "\n";
}


#####################
# User Search
#
# Allows one to select range of users to select, Admin, Supervisors, et cetera

sub UserSearch							# Database, CGI
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter list

	my $table    = new Ethereal::Table($cgi);		# Table handle

	my $radiorange;						# Range radio button
	my $radiofield;						# Field ratio button


	#####################
	# Program area
	
	# Widget sets
	my @range = qw(Administrators Supervisors Banned Verified Non-Priviledged All-Encompasing);
	my @field = qw(E-Mail Login Name Puppet Homepage);

	$radiorange = $cgi->radio_group('RANGE', \@range, $range[5], 'true');
	$radiofield = $cgi->radio_group('FIELD', \@field, $field[0], 'true');


	# Form start	
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), 'SEARCH', 'True'), "\n";


	# Table top
	$table->MakeTop();

	# Search Range
	$table->MakeValid('<B>Search Range</B>', $radiorange);
	$table->MakeBlank();

	# Field and keyword
	$table->MakeValid('<B>Search Field</B>', $radiofield);
	$table->MakeValid('<B>Search Words</B>', $cgi->textfield('KEYWORD', '', 25, 25));	
	$table->MakeBlank();

	# Submit and bottom
	$table->MakeValid($cgi->submit('Search'), ''); 
	$table->MakeBottom();

	# Ending form
	print $cgi->end_form(), "\n";
}


#####################
# User Select
#
# Displays initial query and allows to select individual user

sub UserSelect							# Database, CGI, Parameter List
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter list reference

	my $data     = $database->{'HANDLE'};			# Database handle
	my $query    = SqlHandler($param);			# SQL Query
	my $statement;						# Statement shift

	my $page;						# Homepage handler
	my $url      = $cgi->url(-absolute=>1);			# Return URL

	my $color    = 0;					# Colour changer
	my $home;						# Home page
	my $inline;						# Parameter 

	my @values;						# Received values
	my %home;						# Home page hash


	#####################
	# Program area	
	
	# If defined, remove	
	if (defined($param->{'DELETE'}))
	{
		# Return to orginal condition
		my $login = $param->{'DELETE'};

		# Link to hash
		$database->GetHashHome(\%home);


		# Create instance
		$page = new Ethereal::Home($database, $home{'SetRoot'});


		# Home page handling
		# Retreive home
		($home) = $database->DataGetPuppeteerHome($login);

		# Remove folder
		if (defined($home)) 
		{ $page->DirClear($home); }	


		# Database handling
		# Remove Puppets
		$statement = $data->prepare("SELECT PuppetName 
			FROM  Puppet 
			WHERE PuppeteerLogin=?");
		$statement->execute($login);

		# Clense database of puppets
		while ($_ = $statement->fetchrow())
		{
			$data->do("DELETE FROM PuppetInvitation WHERE PuppetName=?", {}, $_);
			$data->do("DELETE FROM PuppetIgnore WHERE PuppetName=?", {}, $_);
		}

		# Remove puppeteer from tables
		$data->do("DELETE FROM Mail WHERE PuppeteerLogin=?", {}, $login);
		$data->do("DELETE FROM RealmSupervisor WHERE PuppeteerLogin=?", {}, $login);
		$data->do("DELETE FROM PuppetIgnore WHERE PuppeteerLogin=?", {}, $login);
		$data->do("DELETE FROM PuppeteerRegistration WHERE PuppeteerLogin=?", {}, $login);
		$data->do("DELETE FROM PuppeteerRestriction WHERE PuppeteerLogin=?", {}, $login);
		$data->do("DELETE FROM Puppet WHERE PuppeteerLogin=?", {}, $login);
		$data->do("DELETE FROM Puppeteer WHERE PuppeteerLogin=?", {}, $login);

		# Finish statement
		$statement->finish();

		# Cleanup
		$param->Cleanup('DELETE');
	}

	# Back button
	# Generate inline
	($inline) = $param->EmbedInline('USER', $param->{'USER'}, 'CRYPT', $param->{'CRYPT'});

	# Display
	print "<CENTER><A HREF=\"$url$inline\">[Back]</A></CENTER>";	


	# Data gathering and display
	# Prepare and execute
	$statement = $data->prepare($query);
	$statement->execute();


	print "<TABLE WIDTH=\"100%\">";

	# Display header
	print "\t<TR>\n";
	print "\t\t<TD><B>Login Name</B></TD>\n";
	print "\t\t<TD><B>Full Name</B></TD>\n";
	print "\t\t<TD><B>E-Mail address</B></TD>\n";
	print "\t\t<TD></TD>\n";
	print "\t\t<TD></TD>\n";
	print "\t</TR>\n";


	# Fetch rows of data
	while (@values = $statement->fetchrow())
	{
		# Determin colour of row
		if ($color != 0)
		{ print "\t<TR>\n"; $color = 0; }
		else
		{ print "\t<TR BGCOLOR=\"\#999999\">\n";  $color = 1; }

		print "\t\t<TD><B>$values[0]</B></TD>\n";
		print "\t\t<TD>$values[1]</TD>\n";
		print "\t\t<TD><A HREF=\"mailto:$values[2]\">$values[2]</A></TD>\n";


		# Modify button
		# Start form
		print "\t\t", $cgi->start_form();
		print "\t\t", $param->EmbedNormal($param->Flat(), 'LOGIN', $values[0]), "\n";

		print "\t\t<TD>\n";
		print "\t\t\t", $cgi->submit('Modify'), "\n";
		print "\t\t</TD>\n";

		# End form
		print "\t\t", $cgi->end_form(), "\n";


		# Remove button
		# Start form
		print "\t\t", $cgi->start_form();
		print "\t\t", $param->EmbedNormal($param->Flat(), 'DELETE', $values[0]), "\n";

		print "\t\t<TD>\n";
		print "\t\t\t", $cgi->submit('Remove'), "\n";
		print "\t\t</TD>\n";

		# End form
		print "\t\t", $cgi->end_form(), "\n";

		# End row
		print "\t</TR>\n";
	}

	# Finish statement
	$statement->finish();

	# End of table
	print "</TABLE>\n";
}


#####################
# Weblog Handler
#
# The administrative script that administrators will use in order to manage weblog
# entries directly.   This also serves as a way to add entries into the weblog
# itself.

sub WeblogHandler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI filehandle
	my $param    = shift;					# Hash of parameters					

	my $audit;						# Auditing handle
	my $filter;						# Filter handle
	my $date;						# Date handler
	my $table    = shift;					# Table handler
	my $time;						# Current time

	my $data     = $database->{'HANDLE'};			# Direct handle
	my $res;						# Results handle
	my $statement;						# Database statement

	my $nil      = '--NIL--';				# Default
	my $topic;						# Topic widget
	my @topics;						# Topic array

	my %board;						# Board hash
	my %system;						# Board hash


	#####################
	# Program area

	# Define instances
	$audit  = new Ethereal::Audit($database, $cgi, $param);
	$filter = new Ethereal::Filter();
	$date   = new Ethereal::Date();

	# Link with hash
	$database->GetHashBoard(\%board);
	$database->GetHashSystem(\%system);


	# Generate widget
	# Pull values
	$database->GetListTopics(\@topics, $system{'SetBoardAlias'});

	# Append
	unshift(@topics, $nil);

	# Generate
	$topic = $cgi->popup_menu('CLONE', \@topics, $topics[0]);

	# Addition
	# If record has been triggered
	if (defined($param->{'Record'}))
	{
		# Insertion Values
		my ($puppet) = $database->DataGetDefault($param->{'USER'});

		# Post to actually handle
		if ((defined($param->{'POST'}))
		 && (length($param->{'POST'}) > 3)
		 && (length($param->{'SUBJ'}) > 3))
		{
			# Ensure definition
			unless ((defined($param->{'TOPC'}))
			 && (length($param->{'TOPC'}) > 3))
			{
				# Set default topic
				$param->{'TOPC'} = $board{'TxtGeneral'};
			}	

			# Skip if desired
			unless (defined($param->{'DISABLE'}))
			{
				# Post mangling
				# Standard filter
				$param->{'POST'} = $filter->SafeHTML($param->{'POST'});
				$param->{'POST'} = $filter->MakeSane($param->{'POST'}, $board{'SetTruncate'});
			
				# Replace carriage returns
				$param->{'POST'} =~ s/\n/<BR>/gs;

				# Create hyperlinks
				$param->{'POST'} =~ s/(http:\/\/.*?)(\s|$)/<A HREF="$1" TARGET="_blank">$board{TagHyperlink}<\/A> /gm;
				$param->{'POST'} =~ s/([\w\.\-]*?\@[\w\.\-]*?\..*?)(\s|$)/<A HREF="mailto:$1">$board{TagEmail}<\/A> /gm;

				# Strip HTML
				$param->{'SUBJ'} =~ s/<[^>]*>//gs;
				$param->{'TOPC'} =~ s/<[^>]*>//gs;
			}

			# Database updates
			$data->do("INSERT INTO Weblog
				(PuppetName,
				 LogTimestamp,
				 LogTitle,
				 LogTopic,
				 LogPost)
				VALUES (?,?,?,?,?)", {},

			 $puppet,
			 time,
			 $param->{'SUBJ'},
			 lc($param->{'TOPC'}),
			 $param->{'POST'}
			);

			# Audit entry
			$audit->WeblogAdd($param->{'SUBJ'});


			# Cloning vat
			# Only if needed
			if ((defined($param->{'CLONE'}))
			 && ($param->{'CLONE'} ne $nil))
			{


				# Database insertion
				$data->do("INSERT INTO Board
       					(BoardIDNumber,
					 PuppetName,
 					 RealmName,
 					 BoardTopic,
 					 BoardSubject,
 					 BoardTimestamp,
 					 BoardMessage,
 					 BoardUpdate,
					 BoardLevel,
 					 LevelZero,
 					 LevelOne,
 					 LevelTwo,
 					 LevelThree,
 					 LevelFour,
	 				 LevelFive)
        				VALUES (NULL,?,?,?,?,?,?,?,0,0,0,0,0,0,0)", {},

 				$puppet,
 				$system{'SetBoardAlias'},
 				$param->{'CLONE'},
 				$param->{'SUBJ'},
 				time,
 				$param->{'POST'},
 				time);

				# MySQL kludge
				$data->do("UPDATE Board
					SET LevelZero=LAST_INSERT_ID(),
					    LevelOne=LAST_INSERT_ID(),
					    LevelTwo=LAST_INSERT_ID(),
					    LevelThree=LAST_INSERT_ID(),
					    LevelFour=LAST_INSERT_ID(),
					    LevelFive=LAST_INSERT_ID()
					WHERE BoardIDNumber=LAST_INSERT_ID()", {});

			}
		}


		# Cleanup
		$param->Cleanup('Record', 'DISABLE');
	}



	# Display of weblog
	# Prepare and execute
	$statement = $data->prepare("SELECT
		  LogIDNumber,
		  PuppetName,
		  LogTitle,
		  LogTopic,
		  LogTimestamp
		FROM Weblog
		ORDER BY LogIDNumber DESC");
	$statement->execute();


	# Form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat());

	# Table top
	$table->MakeTop();
	$table->MakeSingle('<FONT SIZE=3><B>Weblog contents</B></FONT>');


	# Loop and display
	while ($res = $statement->fetchrow_hashref())
	{
		# Remove
		# No display
		if (defined($param->{"ID$res->{LogIDNumber}"}))
		{
			# Remove
			$data->do("DELETE FROM Weblog
				WHERE LogIDNumber=?", {},

			 $res->{'LogIDNumber'});

			# Cleanup
			$param->Cleanup("ID$res->{LogIDNumber}");

			# Audit entry
			$audit->WeblogRem($res->{'LogIDNumber'}, $res->{'LogTitle'});
		}

		# Display
		else
		{
			# Time setup
			$time = $date->GetDate($res->{'LogTimestamp'});

			# Topic
			$res->{'LogTopic'} = uc($res->{'LogTopic'});

			# Merge and show
			$table->MakeValid(
				"<B>$res->{LogTopic}</B> ($time): $res->{LogTitle}",
				$cgi->checkbox(-name=>"ID$res->{LogIDNumber}", -label=>'')
			 );
		}
	}

	# End query
	$statement->finish();


	# Remove widget
	$table->MakeBlank();
	$table->MakeValid('&nbsp;', $cgi->submit('Purge'));

	# Spacer
	$table->MakeBlank();
	$table->MakeBlank();
	$table->MakeBlank();


	# Addition
	# Title
	$table->MakeSingle('<FONT SIZE=3><B>Addition</B></FONT>');

	# Creation widgets
	$table->MakeValid('<B>Topic</B>', $cgi->textfield('TOPC', '', 15, 25));
	$table->MakeValid('<B>Subject</B>', $cgi->textfield('SUBJ', '', 25, 100));
	$table->MakeBlank();

	# Clone
	$table->MakeValid('<B>Clone</B>', $topic);
	$table->MakeValid('<B>No Filter</B>', $cgi->checkbox(-name=>'DISABLE', -label=>''));
	$table->MakeBlank();

	# Text area
	$table->MakeBlank();
	$table->MakeSingle('<B>Post</B>');
	$table->MakeSingle($cgi->textarea(-name=>'POST',-rows=>10,-columns=>60,-wrap=>'SOFT'));

	# Submit
	$table->MakeBlank();
	$table->MakeSingle($cgi->submit('Record'));
	$table->MakeBottom();


	# Print form
	print $cgi->end_form();

}
