#!/bin/perl
#################################################################################
# Created       : Martin Foster
# Modified      : 22/10/2002
#################################################################################
#
# Gummies  - Script part of Webchat designed to view gummies 
# rights.  Contains comprehensive search parameter.
# Copyright (C) 2000-2002  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 204, 817 5TH ST NE
#                 Calgary, Alberta
#                 T2E 3W9
#
#################################################################################
use CGI qw(:standard);						# Common gateway interface
use CGI::Carp qw(fatalsToBrowser);				# CGI Error logs
use strict;							# Strict variable enforcement

use Ethereal::Database;						# Database handler
use Ethereal::Table;						# Table control
use Ethereal::Template;						# Template handler


#################################################################################
# Data Members
#################################################################################
my $cgi;							# Common gateway interface handle
my $database;							# Database handle


#################################################################################
# Program Area
#################################################################################

	# Initial handles
	$cgi      = new CGI;
	$database = new Ethereal::Database();

	# Connect and fetch
	$database->Connect();


	# Document header
	print $cgi->header();
	$database->DocumentGetHeader();

	# Display gummy
	GummyDisplay($database, $cgi);

	# Document footer
	$database->DocumentGetFooter();


#################################################################################
# Sub-Routines
#################################################################################

#####################
# Gummy Display
#
# Very simple sub-routine that will simply allow the user to look through the 
# list of gummies to search for.

sub GummyDisplay
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle

	my $data     = $database->{'HANDLE'};			# Database handle
	my $tmpl;						# Template handle

	my $res;						# Results			
	my $statement;						# Database statement

	my $class;						# Gummy class
	my $count;						# Counter
	my $ident;						# Gummy indentier
	my $title;						# Display title

	my $fst_name;						# First stored name
	my $fst_path;						# First stored path
	my $sec_name;						# Second stored name
	my $sec_path;						# Second stored path

	my $group;						# Gummy grouping 
	my $realm    = 'system';				# Realm search


	my @classes;						# Gummy classes
	my %config;						# Configuration hash


	#####################
	# Program area

	# Link with has
	$database->GetHashConfiguration(\%config);

	# Create instance
	$tmpl = new Ethereal::Template(\%config);


	# Pull list of gummies
	$database->GetList(\@classes, "SELECT GummyClass
			FROM Gummy
			WHERE GummyClass <> 'default'
			GROUP BY GummyClass
			ORDER BY GummyClass");

	# Display title
	$tmpl->ShowMain($config{TagGummies});

	# Table Top
	print "<CENTER>\n<TABLE WIDTH=500>\n";


	# Classes
	foreach $class (@classes)
	{
		# Query management
		# Prepare and execute
		$statement = $data->prepare("SELECT 
			 GummyName,
			 GummyPath,
			 GummyIdent
			FROM  Gummy
			WHERE GummyClass=?
			  AND RealmName=?
			ORDER BY GummyIdent, GummyName");
		$statement->execute($class, $realm);

		# Reset values
		$count = 0;
		$group = 'None';


		# Class
		$class = ucfirst($class);

		# Title
		$title = $tmpl->PassTitle($class);

		# Display title
		print " <TR>\n",
		      "  <TD COLSPAN=7 HEIGHT=15>&nbsp;</TD>\n",
		      " </TR>\n";
		print " <TR>\n",
		      "  <TD COLSPAN=7>$title</TD>\n",
		      " </TR>\n";


		# Cycle through gummies
		while ($res = $statement->fetchrow_hashref())
		{
			# Increment counter
			$count++;


			# Padding
			$ident = (defined($res->{'GummyIdent'})) ? $res->{'GummyIdent'} : 'None';

			# Title
			if ($ident ne $group)
			{
				# Display single if need be
				if (($count % 2) == 0)
				{
					# Display to screen
					print " <TR>\n",
					      "  <TD WIDTH=10>&nbsp;</TD>\n",
					      "  <TD><B>$fst_name</B></TD>\n",
					      "  <TD><IMG SRC=\"$fst_path\"></TD>\n",
					      "  <TD WIDTH=50 COLSPAN=3>&nbsp;</TD>\n",
					      "  <TD WIDTH=10>&nbsp;</TD>\n",
					      " </TR>\n";
				}	

				# Reset counter
				$count = 1;


				# Prepare for new title
				# Alternate
				$group = $ident;

				# Uppercase
				$ident = ucfirst($ident);

				# Generate title
				$title = $tmpl->PassSub($ident);


				# Display title
				print " <TR>\n",
				      "  <TD WIDTH=10>&nbsp;</TD>\n",
				      "  <TD COLSPAN=5 HEIGHT=15>&nbsp;</TD>\n",
				      "  <TD WIDTH=10>&nbsp;</TD>\n",
				      " </TR>\n";
				print " <TR>\n",
				      "  <TD WIDTH=10>&nbsp;</TD>\n",
				      "  <TD COLSPAN=5>$title</TD>\n",
				      "  <TD WIDTH=10>&nbsp;</TD>\n",
				      " </TR>\n";
			}


			# Store for next pass
			if (($count % 2) != 0)
			{
				# Assign values
				$fst_name = $res->{GummyName};
				$fst_path = $res->{GummyPath};
			}

			# Display row
			else
			{
				# Assign values
				$sec_name = $res->{GummyName};
				$sec_path = $res->{GummyPath};
			

				# Display to screen
				print " <TR>\n",
				      "  <TD WIDTH=10>&nbsp;</TD>\n",
				      "  <TD><B>$fst_name</B></TD>\n",
				      "  <TD><IMG SRC=\"$fst_path\"></TD>\n",
				      "  <TD WIDTH=50>&nbsp;</TD>\n",
				      "  <TD><B>$sec_name</B></TD>\n",
				      "  <TD><IMG SRC=\"$sec_path\"></TD>\n",
				      "  <TD WIDTH=10>&nbsp;</TD>\n",
				      " </TR>\n";
			}
		}
	}

	# End table
	print "</TABLE>\n<CENTER>\n";
}
