#!/bin/perl
################################################################################
# Created       : Martin Foster
# Modified      : 17/11/2002
################################################################################
#
# Private - Script part of Webchat designed to send private messages
# Copyright (C) 2000-2002  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 204, 817 5TH ST NE
#                 Calgary, Alberta
#                 T2E 3W9
#
################################################################################

use CGI qw(:standard);						# Common gateway interface
use CGI::Carp qw(fatalsToBrowser);				# CGI Error logs
use strict;							# Strict variable enforcement

use Ethereal::Comm;						# Communication handler
use Ethereal::Database;						# Database handler
use Ethereal::Login;						# Login functionality
use Ethereal::Option;						# Option handling
use Ethereal::Param;						# Parameter control
use Ethereal::Table;						# Table handler


################################################################################
# Data Members
################################################################################
my $cgi;							# Common gateway interface handle
my $comm;							# Communications panel
my $database;							# Database handle
my $login;							# Login handle
my $param;							# Parameter handle

my $access;							# Access type
my $count;							# Amount of targets
my $post;							# Temporary post
my $target;							# Individual targets
my @targets;							# Targets to send to

my %realm;							# Realm DBM


################################################################################
# Program Area
################################################################################

	# Initial handles
	$cgi      = new CGI;
	$database = new Ethereal::Database();

	# Fetch target
	@targets = $cgi->param('TARGET');

	# Connect and fetch
	$database->Connect();

	# Parameter handling
	$param = new Ethereal::Param($database, $cgi);
	$param->GetParam();

	# Fetch hash
	$database->GetHashRealm(\%realm);


	# Login and Comm handles
	$login = new Ethereal::Login($database, $cgi, $param);
	$comm  = new Ethereal::Comm($database, $cgi);

	# Authenthication
	if ($login->GetVerificationRealm())
	{
		# Prevent bans
		$login->CheckLegality(\%realm);

		# Only do the following if posting is needed
		if (defined($param->{'PSEND'}))
		{
			# Pull values
			($access) = $database->DataGetRealmAccess($param->{'ROOM'});


			# Only if target is valid
			if ($param->{'TARGET'} ne $realm{'TxtNone'})
			{
				# Retreive count
				$count = @targets;

				# Determin post
				$post = ($count > 1) ? "$realm{TagMulti} $param->{POST}" : "$param->{POST}";
			
				# Circle through hosts
				foreach $target (@targets)
				{
					# Set value of target
					$param->{'TARGET'} = $target;
					$param->{'POST'}   = $post;
		
					TargetHandler($database, $param, \%realm, $access);
					$comm->DataWrite($param);
				}
			}
		}

		# Display panel
		DisplayHandler($database, $cgi, $param, $comm, \%realm, $access);
	}


################################################################################
# Sub-Routines
################################################################################

####################
# Display Handler
#
# Display prompt for a private message, this includes a list of all the puppets
# et cetera. 

sub DisplayHandler						# Database, Cgi, Param
{
	####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# Cgi handle
	my $param    = shift;					# Parameter handle
	my $comm     = shift;					# Communications panel
	my $realm    = shift;					# Realm options
	my $access   = shift;					# Realm access

	my $option   = new Ethereal::Option();			# Option handle

	my $access;						# Access control
	my $admin;						# Admin priviliges
	my $js;							# Javascript control

	my $defcolor;						# Default colour
	my $defface;						# Default face
	my $defsize;						# Default size
	my $delay;						# Delay of time
	my $panel;						# Private message panel

	my @who;						# List of puppets
	my @char;						# List of characters
	my @color;						# Colour choices
	my @face;						# Font face
	my @size;						# Text colour



	####################
	# Program area

	# Pull values
	($admin) = $database->DataGetAdmin($param->{'USER'});
	($js) = $database->DataGetJavascript($param->{'USER'});

	# Calculate delay
	$delay = (time - $realm->{'SetTimeout'});


	# Populate arrays
        @color = $comm->FontSelectColor($param);
        @size  = $comm->FontSelectSize($param);
        @face  = $comm->FontSelectFace($param);

	# Populate puppet list
	if ($access ne 'restricted')
	{
		# Pull list of all applicable handles
		$database->GetListPuppetRealm(\@char, $param->{'USER'}, $param->{'ROOM'});
		$database->GetListWho(\@who, $param->{'ROOM'}, $delay);
	}
	else
	{
		# Pull list of canon handles
		$database->GetListCanonPuppets(\@char, $param->{'USER'}, $param->{'ROOM'});
		$database->GetListCanonWho(\@who, $param->{'ROOM'}, $delay);
	}

	# Sort colours
	@color = sort(@color);

	
	# Retreive size and color
	($defsize, $defcolor, $defface) = $database->DataGetFont($param->{'CHAR'}, $param->{'USER'}, $color[0], $face[0]);

	# Allow for system messages
	push(@who, $realm->{'TxtSystem'}) if ($admin eq 'yes'); 

	# Default selection
	unshift(@who, $realm->{'TxtNone'});


	# Targets to who
	$target = (defined($param->{'OTARGET'})) ? $param->{'OTARGET'} : $who[0];

	# Cleanup for embed
	$param->Cleanup('POST');


	# Retreive private message panel
	$panel = $database->FetchRealmPrivate($param->{'ROOM'});


	# Javascript controls
	if ($js eq 'yes')
	{
		# Pull body tag
		if ($panel =~ /<(BODY.*?)>/si)
		{
			# Assign
			my $org = $1;
			my $new = $1;

			# Prepare
			$new =~ s/(\W)/\\$1/gs;
			

			# Replace
		 	$panel =~ s/$new/$org onload="document.SEND.POST.focus();"/gs;	
		}
	}


	# Widgets
	# Assigning values
	my $wpostsend = $cgi->submit($realm{'TxtSubmitButton'});
	my $wposttext = $cgi->textarea(-name=>'POST',-rows=>3,-columns=>55,-wrap=>'SOFT');
	my $wchartrgt = $cgi->scrolling_list('TARGET', \@who, $target, 5, 'true');
	my $wcharslct = $cgi->popup_menu('CHAR', \@char, $param->{'CHAR'});
	my $wfontcolr = $cgi->popup_menu('COLOR', \@color, $defcolor);
	my $wfontface = $cgi->popup_menu('FACE', \@face, $defface);
	my $wfontsize = $cgi->popup_menu('SIZE', \@size, $defsize);

	# Form values
	my $wformstrt = $cgi->start_form(-name=>'SEND');
	my $wformhidn = $param->EmbedNormal($param->Flat(), 'PSEND', 'True');
	my $wformends = $cgi->end_form();


	# Widget replacements
	# Posting widgets
	$panel =~ s/WPOSTSEND/$wpostsend/gs;
	$panel =~ s/WPOSTTEXT/$wposttext/gs;

	# Puppet selection
	$panel =~ s/WCHARSLCT/$wcharslct/gs;
	$panel =~ s/WCHARTRGT/$wchartrgt/gs;

	# Font selection
	$panel =~ s/WFONTCOLR/$wfontcolr/gs;
	$panel =~ s/WFONTFACE/$wfontface/gs;
	$panel =~ s/WFONTSIZE/$wfontsize/gs;

	# Form
	$panel =~ s/WFORMSTRT/$wformstrt$wformhidn/gs;
	$panel =~ s/WFORMENDS/$wformends/gs;


	# Display HTML
	# Document header
	print $cgi->header;

	# Document
	print "$panel";
}

#####################
# Target Handler
#
# Will determine target to send the message privately to.  This will also create a
# parameter called SOURCE which will receive a copy as well.

sub TargetHandler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $param    = shift;					# Parameter handle
	my $realm    = shift;					# Chat options
	my $access   = shift;					# Realm access

	my $puppeteer;						# Puppeteer to send to


	#####################
	# Program area

	# Only do changes if target exists
	if (defined($param->{'TARGET'}))
	{
		# Retreive puppeteer name
		$param->{'OTARGET'} = $param->{'TARGET'};	# Keep original value

		# Only retreive if needed
		if (($param->{'OTARGET'} ne 'dice roll') 
		 && ($param->{'OTARGET'} ne $realm->{'TxtSystem'}))
		{ 
			# Retreive
			# Determin query to run
			if ($access ne 'restricted')
			{	
				# System level handles
				($puppeteer) = $database->DataGetPuppeteerLogin($param->{'TARGET'}); 
			}
			else
			{
				# Source level handles
				($puppeteer) = $database->DataGetPuppeteerCanon($param->{'ROOM'}, $param->{'TARGET'}); 
			}

			# Assign target
			$param->{'TARGET'} = $puppeteer;
		}

		# Assign values to parameters
		$param->{'SOURCE'} = $param->{'USER'};
	}
}
